/* Dammit device driver for Linux
   At the moment, it supports the following devices:
   /dev/dammit0 u 253 0  8 bit dammit on data pins.
   /dev/dammit1 u 253 1  12 bit dammit on data + control pins (Untested, but should work.)
   /dev/dammit2 u 253 2  64 bit multiplexed dammit (Completely untested. May need timing fixes.)
*/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/wrapper.h>
#include <linux/parport.h>

#include "dammit.h"

/* In 2.2.3 /usr/include/linux/version.h includes 
 * a macro for this, but 2.0.35 doesn't - so I add 
 * it here if necessary. */
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif


/* Conditional compilation. LINUX_VERSION_CODE is 
 * the code (as per KERNEL_VERSION) of this version. */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h> /* for put_user */
#endif

#define SUCCESS 0


/* Deal with CONFIG_MODVERSIONS */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif


/* Device Declarations **************************** */

/* parport handle */
struct parport *myport = NULL;
struct pardevice *mypdevice = NULL;

/* Is the device open right now? Used to prevent 
 * concurent access into the same device */
static int Device_Open = 0;
/* Minor number of currently open device */
static int Minor;

/* Bitmasks used to mask the final output */
static unsigned char mask_on = 0x00; /* Any 1's in here are always on */
static unsigned char mask_off = 0x00; /* Any 1's in here are never on */

/* The number of channels we're running on the multiplexer. */
static unsigned int multi_count = 1;

/* This function is called whenever a process 
 * attempts to open the device file */
static int device_open(struct inode *inode, 
                       struct file *file)
{
  /* We don't want to talk to two processes at the 
   * same time */
  if (Device_Open)
    return -EBUSY;
  if( parport_claim( mypdevice ) ) {
	return -EBUSY;
  }
  /* If this was a process, we would have had to be 
   * more careful here.
   *
   * In the case of processes, the danger would be 
   * that one process might have check Device_Open 
   * and then be replaced by the schedualer by another 
   * process which runs this function. Then, when the 
   * first process was back on the CPU, it would assume 
   * the device is still not open. 
   *
   * However, Linux guarantees that a process won't be 
   * replaced while it is running in kernel context. 
   *
   * In the case of SMP, one CPU might increment 
   * Device_Open while another CPU is here, right after 
   * the check. However, in version 2.0 of the 
   * kernel this is not a problem because there's a lock 
   * to guarantee only one CPU will be kernel module at 
   * the same time. This is bad in  terms of 
   * performance, so version 2.2 changed it. 
   * Unfortunately, I don't have access to an SMP box 
   * to check how it works with SMP.
   */

  Device_Open++;
  Minor = (inode->i_rdev & 0xFF); 
  if( Minor > 2 ) /* only one we support right now */
	return -EINVAL;

  /* Make sure that the module isn't removed while 
   * the file is open by incrementing the usage count 
   * (the number of opened references to the module, if 
   * it's not zero rmmod will fail)
   */
  MOD_INC_USE_COUNT;

  return SUCCESS;
}


/* This function is called when a process closes the 
 * device file. It doesn't have a return value in 
 * version 2.0.x because it can't fail (you must ALWAYS
 * be able to close a device). In version 2.2.x it is 
 * allowed to fail - but we won't let it. 
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int device_release(struct inode *inode, 
                          struct file *file)
#else 
static void device_release(struct inode *inode, 
                           struct file *file)
#endif
{
  parport_release( mypdevice );

  /* We're now ready for our next caller */
  Device_Open --;

  /* Decrement the usage count, otherwise once you 
   * opened the file you'll never get rid of the module.
   */
  MOD_DEC_USE_COUNT;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  return 0;
#endif
}


/* This function is called whenever a process which 
 * have already opened the device file attempts to 
 * read from it. */


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t device_read(struct file *file,
    char *buffer,    /* The buffer to fill with data */
    size_t length,   /* The length of the buffer */
    loff_t *offset)  /* Our offset in the file */
#else
static int device_read(struct inode *inode,
                       struct file *file,
    char *buffer,   /* The buffer to fill with 
                     * the data */ 
    int length)     /* The length of the buffer 
                     * (mustn't write beyond that!) */
#endif
{
  return -EINVAL;
}




/* This function is called when somebody tries to write 
 * into our device file */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t device_write(struct file *file,
    const char *buffer,    /* The buffer */
    size_t length,   /* The length of the buffer */
    loff_t *offset)  /* Our offset in the file */
#else
static int device_write(struct inode *inode,
                        struct file *file,
                        const char *buffer,
                        int length)
#endif
{
  
  /* This switches based on which dammit box we're using:
     0: 8 bit dammit (Data only)
     1: 12 bit dammy (Data + control hooked to lights)
     2: MegaDammit (Multiple 8 bit banks)
  */
  switch( Minor ) {
	
	/*8 Bit dammit, one light on each data pin. */
      case DAMMIT_8:
      { 
	  unsigned char mask;
  	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0) 
     	  get_user( mask, buffer );
  	#else
     	  mask = get_user( buffer );
  	#endif
	  /* Turn on any set in the mask */
	  mask |= mask_on;
	  /* Turn off any set in the mask */
          mask &= ~mask_off;
  	  parport_write_data( myport, mask ); 
  	return 1;
      }
	/* 12 Bit dammit running 4 lights off the control pins. */
      case DAMMIT_12:
      { 
	static unsigned char block[2];
	static unsigned int pos = 0;
	int len = length;
	while( len > 0 ) {
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
		  get_user( block[pos], buffer++ );
		#else
	  	  block[pos] = get_user( buffer );
		#endif
		pos++;
		len--;
		if( pos == 2 ) {
	/* Need to flip the control bits, and mask off the upper 4. */
			block[1] ^= 0x0B;
			block[1] &= 0x0F;
			parport_write_data( myport, block[0] );
			parport_write_control( myport, block[1]);
			pos = 0;
		}
		return length;
	}
     }
     case DAMMIT_MULTI:
     {
	  static unsigned char buf[8];
	  static unsigned char pos = 0;
	  int len = length;
	  while( len > 0 ) {
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
		  get_user( buf[pos++], buffer++ );
		#else
		  buf[pos++] = get_user( buffer++ );
		#endif
		len--;
		if( pos == multi_count ) {
			for( pos = 0; pos < multi_count; pos++ ) {
				int control_mask = ((pos << 1) ^ 0x0A); 
				parport_write_control( myport,  control_mask );
				parport_write_data( myport, buf[pos] );
				parport_write_control( myport, control_mask | 0x01 );
			}
			pos = 0;
		}
	  }
	return length;
      } 
     default:
          return -EINVAL;  /* Any others we just return an error. */
  }
}

/* ioctl */
int device_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
  switch( ioctl_num ) {
	case DAMMIT_SET_ON_MASK:
		mask_on = (unsigned char) ioctl_param;
		return 0;
	case DAMMIT_SET_OFF_MASK:
		mask_off = (unsigned char) ioctl_param;
		return 0;
	case DAMMIT_GET_ON_MASK:
		return mask_on;
	case DAMMIT_GET_OFF_MASK:
		return mask_off;
	case DAMMIT_SET_MULTI:
		if( Minor != 2 )
			return -EINVAL;
		if( ( ioctl_param > 8 ) || ( ioctl_param < 1 ) )
			return -EINVAL;
		multi_count = ioctl_param;
		return 0;
   }
   return -EINVAL;
}
/* Module Declarations ***************************** */
static int Major;

struct file_operations Fops = {
  NULL,   /* seek */
  device_read, 
  device_write,
  NULL,   /* readdir */
  NULL,   /* select */
  device_ioctl,
  NULL,   /* mmap */
  device_open,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  NULL,   /* flush */
#endif
  device_release  /* a.k.a. close */
};


/* Initialize the module - Register the character device */
int init_module()
{
  /* Register the character device (atleast try) */
  Major = module_register_chrdev(DAMMIT_MAJOR, 
                                 DAMMIT_NAME,
                                 &Fops);

  /* Negative values signify an error */
  if (Major < 0) {
    printk ("%s device failed with %d\n",
            "Sorry, registering the character",
            Major);
    return Major;
  }

  myport = parport_enumerate();
  if( myport == NULL ) {
	printk( "dammit couldn't get a parallel port\n" );
	return -1;
  }
 
  mypdevice = parport_register_device( myport, "dammit", NULL, NULL, NULL, 0,
	NULL );
  if( mypdevice == NULL ) {
	printk( "dammit couldn't grab a pardevice\n" );
	return -1;
  }
  printk ("Dammit registerd on major device number %d, parallel port at 0x%x\n",
          DAMMIT_MAJOR, (unsigned int)myport->base);
 
  return 0;
}


/* Cleanup - unregister the appropriate file from /proc */
void cleanup_module()
{
  int ret;
  
  if( mypdevice ) {
	parport_unregister_device( mypdevice );
  }

  /* Unregister the device */
  ret = module_unregister_chrdev(DAMMIT_MAJOR, DAMMIT_NAME);
 
  /* If there's an error, report it */ 
  if (ret < 0)
    printk("Error in unregister_chrdev: %d\n", ret);
}
